/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: phk_argv.c,v 1.3 2006/09/09 13:51:39 phk Exp $
 *
 */

#include <ctype.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "phk_argv.h"

static char enomem[] = "Memory allocation error";
static char *nomem[2] = {enomem, NULL};
static char eillbs[] = "Illegal backslash sequence";
static char *illbs[2] = {eillbs, NULL};
static char emissq[] = "Missing '\"'";
static char *missq[2] = {emissq, NULL};

static int
BackSlash(const char *s, int *res)
{
	int i, r;

	assert(*s == '\\');
	r = i = 0;
	switch(s[1]) {
	case 'n':
		i = '\n';
		r = 2;
		break;
	case 'r':
		i = '\r';
		r = 2;
		break;
	case 't':
		i = '\t';
		r = 2;
		break;
	case '"':
		i = '"';
		r = 2;
		break;
	case '\\':
		i = '\\';
		r = 2;
		break;
	case '0': case '1': case '2': case 3:
	case '4': case '5': case '6': case 7:
		for (r = 1; r < 4; r++) {
			if (!isdigit(s[r]))
				break;
			if (digittoint(s[r]) > 7)
				break;
			i <<= 3;
			i |= digittoint(s[r]);
		}
		break;
	case 'x':
		if (1 == sscanf(s + 1, "x%02x", &i))
			r = 4;
		break;
	default:
		break;
	}
	if (res != NULL)
		*res = i;
	return (r);
}

static char *
BackSlashDecode(const char *s, const char *e)
{
	const char *q;
	char *p, *r;
	int i, j;

	p = calloc((e - s) + 1, 1);
	if (p == NULL)
		return (p);
	for (r = p, q = s; q < e; ) {
		if (*q != '\\') {
			*r++ = *q++;
			continue;
		}
		i = BackSlash(q, &j);
		q += i;
		*r++ = j;
	}
	*r = '\0';
	return (p);
}

char **
PhkArgvParse(const char *s, int comment)
{
	char **argv, **av2;
	const char *p;
	int nargv, largv;
	int i, quote;

	assert(s != NULL);
	nargv = 1;
	largv = 16;
	argv = calloc(sizeof *argv, largv);
	if (argv == NULL)
		return (nomem);

	for (;;) {
		argv[nargv] = NULL;
		if (*s == '\0')
			break;
		if (isspace(*s)) {
			s++;
			continue;
		}
		if (comment && *s == '#')
			break;
		if (*s == '"') {
			p = ++s;
			quote = 1;
		} else {
			p = s;
			quote = 0;
		}
		while (1) {
			if (*s == '\\') {
				i = BackSlash(s, NULL);
				if (i == 0) {
					PhkArgvFree(argv);
					return (illbs);
				}
				s += i;
				continue;
			}
			if (!quote) {
				if (*s == '\0' || isspace(*s))
					break;
				s++;
				continue;
			}
			if (*s == '"')
				break;
			if (*s == '\0') {
				PhkArgvFree(argv);
				return (missq);
			}
			s++;
		}
		if (nargv + 1 >= largv) {
			av2 = realloc(argv, sizeof (*argv) * (largv += largv));
			if (av2 == NULL) {
				PhkArgvFree(argv);
				return (nomem);
			}
			argv = av2;
		}
		argv[nargv] = BackSlashDecode(p, s);
		if (argv[nargv] == NULL) {
			PhkArgvFree(argv);
			return (nomem);
		}
		nargv++;
		if (*s != '\0')
			s++;
	}
	return (argv);
}

void
PhkArgvFree(char **argv)
{
	int i;
	
	if (argv == NULL || argv[0] != NULL)
		return;
	for (i = 1; argv[i] != NULL; i++)
		free(argv[i]);
	free(argv);
}

#ifdef TESTPROG

#include <printf.h>

static void
PrintArgv(char **argv)
{
	int i;

	printf("---- %p\n", argv);
	if (argv[0] != NULL)
		printf("err %V\n", argv[0]);
	for (i = 1; argv[i] != NULL; i++)
		printf("%3d %V\n", i, argv[i]);
}

static void
Test(const char *str)
{
	char **av;

	printf("Test: <%V>\n", str);
	av = ParseArgv(str, 0);
	PrintArgv(av);
}

int
main(int argc __unused, char **argv __unused)
{
	char buf[BUFSIZ];

        register_printf_render_std("V");

	while (fgets(buf, sizeof buf, stdin))
		Test(buf);

	return (0);
}
#endif
