/*
 * $Id: phk_event.c,v 1.2 2007/11/10 19:12:19 phk Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <errno.h>
#include <poll.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#include "phk_event.h"
#include "phk_obj.h"
#include "phk_binheap.h"

/* INFTIM indicates an infinite timeout for poll(2) */
#ifndef INFTIM
#define INFTIM -1
#endif

struct evsig {
	struct phkevbase		*evb;
	struct phkev		*ev;
	struct sigaction	sigact;
	unsigned char		happened;
};

static struct evsig		*ev_sigs;
static int			ev_nsig;

struct phkevbase {
	unsigned		magic;
#define EVBASE_MAGIC		0x0cfd976f
	TAILQ_HEAD(,phkev__)	events;
	struct pollfd		*pfd;
	unsigned		npfd;
	unsigned		lpfd;
	struct phkbinheap	*binheap;
	unsigned char		compact_pfd;
	unsigned char		disturbed;
	unsigned		psig;
};

struct phkev__ {
	unsigned		magic;
#define PHKEV___MAGIC		0x3c518c90
	struct phkev		ev;
	double			when;
	TAILQ_ENTRY(phkev__)	list;
	unsigned		binheap_idx;
	unsigned		privflags;
	struct phkevbase	*evb;
	unsigned		poll_idx;
};

/*--------------------------------------------------------------------*/

static double
ev_now(void)
{
	double t;
	struct timespec ts;

	assert(clock_gettime(CLOCK_MONOTONIC, &ts) == 0);
	t = ts.tv_sec + ts.tv_nsec * 1e-9;
	return (t);
}

/*--------------------------------------------------------------------*/

static void
ev_bh_update(void *priv, void *a, unsigned u)
{
	struct phkevbase *evb;
	struct phkev__ *e;

	CAST_OBJ_NOTNULL(evb, priv, EVBASE_MAGIC);
	CAST_OBJ_NOTNULL(e, a, PHKEV___MAGIC);
	e->binheap_idx = u;
}

static int
ev_bh_cmp(void *priv, void *a, void *b)
{
	struct phkevbase *evb;
	struct phkev__ *ea, *eb;

	CAST_OBJ_NOTNULL(evb, priv, EVBASE_MAGIC);
	CAST_OBJ_NOTNULL(ea, a, PHKEV___MAGIC);
	CAST_OBJ_NOTNULL(eb, b, PHKEV___MAGIC);
	return (ea->when < eb->when);
}

/*--------------------------------------------------------------------*/

static int
ev_get_pfd(struct phkevbase *evb)
{
	unsigned u;
	void *p;

	if (evb->lpfd < evb->npfd) 
		return (0);

	if (evb->npfd > 256)
		u = evb->npfd + 256;
	else if (evb->npfd < 8)
		u = 8;
	else
		u = evb->npfd * 2;
	p = realloc(evb->pfd, sizeof *evb->pfd * u);
	if (p == NULL)
		return (1);
	evb->npfd = u;
	evb->pfd = p;
	return (0);
}

/*--------------------------------------------------------------------*/

static int
ev_get_sig(int sig)
{
	struct evsig *os;

	if (sig < ev_nsig) 
		return (0);

	os = calloc(sizeof *os, (sig + 1));
	if (os == NULL)
		return (ENOMEM);

	memcpy(os, ev_sigs, ev_nsig * sizeof *os);

	free(ev_sigs);
	ev_sigs = os;
	ev_nsig = sig + 1;

	return (0);
}

/*--------------------------------------------------------------------*/

static void
ev_sighandler(int sig)
{
	struct evsig *es;

	assert(sig < ev_nsig);
	assert(ev_sigs != NULL);
	es = &ev_sigs[sig];
	if (!es->happened)
		es->evb->psig++;
	es->happened = 1;
}

/*--------------------------------------------------------------------*/

struct phkevbase *
PhkEvNewBase(void)
{
	struct phkevbase *evb;

	evb = calloc(sizeof *evb, 1);
	if (evb == NULL)
		return (evb);
	if (ev_get_pfd(evb)) {
		free(evb);
		return (NULL);
	}
	evb->magic = EVBASE_MAGIC;
	TAILQ_INIT(&evb->events);
	evb->binheap = PhkBinHeapNew(evb, ev_bh_cmp, ev_bh_update);
	return (evb);
}

/*--------------------------------------------------------------------*/

void
PhkEvDestroyBase(struct phkevbase *evb)
{
	CHECK_OBJ_NOTNULL(evb, EVBASE_MAGIC);
	evb->magic = 0;
	free(evb);
}

/*--------------------------------------------------------------------*/

struct phkev *
PhkEvNew(void)
{
	struct phkev__ *e;

	e = calloc(sizeof *e, 1);
	if (e != NULL) {
		e->magic = PHKEV___MAGIC;
		e->ev.magic = PHKEV_MAGIC;
		e->ev.fd = -1;
		e->ev.real = e;
	}
	return (&e->ev);
}

/*--------------------------------------------------------------------*/

int
PhkEvAdd(struct phkevbase *evb, struct phkev *e)
{
	struct evsig *es;
	struct phkev__ *E;

	CHECK_OBJ_NOTNULL(evb, EVBASE_MAGIC);
	CHECK_OBJ_NOTNULL(e, PHKEV_MAGIC);
	E = e->real;
	CHECK_OBJ_NOTNULL(E, PHKEV___MAGIC);
	assert(e->callback != NULL);
	assert(e->sig >= 0);
	assert(e->timeout >= 0.0);
	assert(e->fd < 0 || e->fd_flags);

	if (e->sig > 0 && ev_get_sig(e->sig))
		return (ENOMEM);

	if (e->fd >= 0 && ev_get_pfd(evb))
		return (ENOMEM);

	if (e->sig > 0) {
		es = &ev_sigs[e->sig];
		if (es->ev != NULL)
			return (EBUSY);
		assert(es->happened == 0);
		es->ev = e;
		es->evb = evb;
		es->sigact.sa_flags = e->sig_flags;
		es->sigact.sa_handler = ev_sighandler;
	} else {
		es = NULL;
	}

	if (e->fd >= 0) {
		assert(evb->lpfd < evb->npfd);
		evb->pfd[evb->lpfd].fd = e->fd;
		evb->pfd[evb->lpfd].events =
		    e->fd_flags & (EV_RD|EV_WR|EV_ERR|EV_HUP);
		E->poll_idx = evb->lpfd;
		evb->lpfd++;
	} else
		E->poll_idx = -1;

	e->magic = PHKEV_MAGIC;	/* before binheap_insert() */

	if (e->timeout != 0.0) {
		E->when += ev_now() + e->timeout;
		PhkBinHeapInsert(evb->binheap, E);
		assert(E->binheap_idx > 0);
	} else {
		E->when = 0.0;
		E->binheap_idx = 0;
	}

	E->evb = evb;
	E->privflags = 0;
	if (e->fd < 0)
		TAILQ_INSERT_TAIL(&evb->events, E, list);
	else
		TAILQ_INSERT_HEAD(&evb->events, E, list);

	if (e->sig > 0) {
		assert(es != NULL);
		assert(sigaction(e->sig, &es->sigact, NULL) == 0);
	}

	return (0);
}

/*--------------------------------------------------------------------*/

void
PhkEvDel(struct phkevbase *evb, struct phkev *e)
{
	struct evsig *es;
	struct phkev__ *E;

	CHECK_OBJ_NOTNULL(evb, EVBASE_MAGIC);
	CHECK_OBJ_NOTNULL(e, PHKEV_MAGIC);
	E = e->real;
	CHECK_OBJ_NOTNULL(E, PHKEV___MAGIC);
	assert(evb == E->evb);

	if (E->binheap_idx != 0)
		PhkBinHeapDelete(evb->binheap, E->binheap_idx);
	assert(E->binheap_idx == 0);

	if (e->fd >= 0) {
		evb->pfd[E->poll_idx].fd = -1;
		if (E->poll_idx == evb->lpfd - 1)
			evb->lpfd--;
		else
			evb->compact_pfd++;
		e->fd = -1;
	}

	if (e->sig > 0) {
		assert(e->sig < ev_nsig);
		es = &ev_sigs[e->sig];
		assert(es->ev == e);
		es->ev = NULL;
		es->evb = NULL;
		es->sigact.sa_flags = e->sig_flags;
		es->sigact.sa_handler = SIG_DFL;
		assert(sigaction(e->sig, &es->sigact, NULL) == 0);
		es->happened = 0;
	}

	TAILQ_REMOVE(&evb->events, E, list);

	E->evb = NULL;

	evb->disturbed = 1;
}

/*--------------------------------------------------------------------*/

int
PhkEvSched(struct phkevbase *evb)
{
	int i;

	CHECK_OBJ_NOTNULL(evb, EVBASE_MAGIC);
	do 
		i = PhkEvSchedOne(evb);
	while (i == 1);
	return (i);
}

/*--------------------------------------------------------------------*/

static void
ev_compact_pfd(struct phkevbase *evb)
{
	unsigned u;
	struct pollfd *p;
	struct phkev__ *EP;

	p = evb->pfd;
	EP = TAILQ_FIRST(&evb->events);
	for (u = 0; u < evb->lpfd; u++, p++) {
		if (p->fd >= 0)
			continue;
		for(; EP != NULL; EP = TAILQ_NEXT(EP, list)) {
			if (EP->ev.fd >= 0 && EP->poll_idx > u)
				break;
		} 
		if (EP == NULL)
			break;
		*p = evb->pfd[EP->poll_idx];
		EP->poll_idx = u;
	}
	evb->lpfd = u;
	evb->compact_pfd = 0;
}

/*--------------------------------------------------------------------*/

static int
ev_sched_timeout(struct phkevbase *evb, struct phkev__ *E, double t)
{
	int i;

	i = E->ev.callback(&E->ev, 0);
	if (i) {
		PhkEvDel(evb, &E->ev);
		free(E);
	} else {
		E->when = t + E->ev.timeout;
		PhkBinHeapDelete(evb->binheap, E->binheap_idx);
		PhkBinHeapInsert(evb->binheap, E);
	}
	return (1);
}

static int
ev_sched_signal(struct phkevbase *evb)
{
	int i, j;
	struct evsig *es;
	struct phkev *e;

	es = ev_sigs;
	for (j = 0; j < ev_nsig; j++, es++) {
		if (!es->happened || es->evb != evb)
			continue;
		evb->psig--;
		es->happened = 0;
		e = es->ev;
		assert(e != NULL);
		i = e->callback(e, EV_SIG);
		if (i) {
			PhkEvDel(evb, e);
			free(e);
		}
	}
	return (1);
}

int
PhkEvSchedOne(struct phkevbase *evb)
{
	double t;
	struct phkev__ *E, *E2, *E3;
	int i, j, tmo;
	struct pollfd *pfd;

	CHECK_OBJ_NOTNULL(evb, EVBASE_MAGIC);
	E = PhkBinHeapRoot(evb->binheap);
	if (E != NULL) {
		CHECK_OBJ_NOTNULL(E, PHKEV___MAGIC);
		assert(E->binheap_idx == 1);
		t = ev_now();
		if (E->when <= t)
			return (ev_sched_timeout(evb, E, t));
		tmo = (E->when - t) * 1e3;
		if (tmo == 0)
			tmo = 1;
	} else
		tmo = INFTIM;

	if (evb->compact_pfd)
		ev_compact_pfd(evb);

	if (tmo == INFTIM && evb->lpfd == 0)
		return (0);

	if (evb->psig)
		return (ev_sched_signal(evb));
	assert(evb->lpfd < evb->npfd);
	i = poll(evb->pfd, evb->lpfd, tmo);
	if(i == -1 && errno == EINTR)
		return (ev_sched_signal(evb));
	if (i == 0) {
		assert(E != NULL);
		t = ev_now();
		if (E->when <= t)
			return (ev_sched_timeout(evb, E, t));
	}
	evb->disturbed = 0;
	TAILQ_FOREACH_SAFE(E, &evb->events, list, E2) {
		if (i == 0)
			break;
		if (E->ev.fd < 0)
			continue;
		assert(E->poll_idx < evb->lpfd);
		pfd = &evb->pfd[E->poll_idx];
		assert(pfd->fd == E->ev.fd);
		if (!pfd->revents)
			continue;
		j = E->ev.callback(&E->ev, pfd->revents);
		i--;
		if (evb->disturbed) {
			TAILQ_FOREACH(E3, &evb->events, list) {
				if (E3 == E) {
					E3 = TAILQ_NEXT(E, list);
					break;
				} else if (E3 == E2)
					break;
			}
			E2 = E3;
			evb->disturbed = 0;
		}
		if (j) {
			PhkEvDel(evb, &E->ev);
			evb->disturbed = 0;
			free(E);
		}
	}
	return (1);
}
