/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: phk_expr.c,v 1.3 2008/04/29 18:50:56 phk Exp $
 *
 * Function to evaluate a string as a floating point expression.
 *
 * Available operations:
 *
 *	+ - * / 	basic arithmetic
 *	^		exponentiation
 *	(...)		priority
 *	- +		unary operators
 *	foo(...)	function (see table below)
 *	[...]		comment
 * 
 * The overall outline is inspired by the LCC book [ISBN 0-8053-1670-1]
 */

#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "phk_expr.h"

/*
 * The state we pass around
 */
struct e {
	const char *b, *p, *e;	/* beginning, pointer and end */
	const char *err;	/* error string */
	int t;			/* token */
	int l;			/* precedence level of token */
};

static double expr(struct e *e, int level);


/*
 * Debug function, enable as needed.
 */

static void
dbg(struct e *e, const char *w)
{
	if (1)
		return;
	fprintf(stderr, "<%*.*s|", e->p - e->b, e->p - e->b, e->b);
	fprintf(stderr, "%*.*s> %s\n", e->e - e->p, e->e - e->p, e->p, w);
}

/*
 * Skip whitespace and comments
 */

static void
skip_space(struct e *e)
{

	while (e->p < e->e && isspace(*e->p))
		e->p++;

	/* We allow for comments enclosed in [...] */
	if (e->p < e->e && *e->p == '[') {
		while (e->p < e->e && *e->p != ']')
			e->p++;
		e->p++;
		skip_space(e);
	}
}

/*
 * Recognize the diadic operation tokens and assign priority
 *
 * We reuse the precedence from the LCC book, even if we do not
 * implement most of them.
 */

static void
get_token(struct e *e)
{

	dbg(e, "token{");
	e->t = 0;
	e->l = 0;
	skip_space(e);
	if (e->p == e->e || e->err != NULL)
		;
	else if (*e->p == '*') { e->t = *e->p; e->p++; e->l = 13; }
	else if (*e->p == '/') { e->t = *e->p; e->p++; e->l = 13; }
	else if (*e->p == '+') { e->t = *e->p; e->p++; e->l = 12; }
	else if (*e->p == '-') { e->t = *e->p; e->p++; e->l = 12; }
	else if (*e->p == '^') { e->t = *e->p; e->p++; e->l = 14; }
	dbg(e, "token}");
}

/*
 * Private math functions
 */
static double
mylog2(double a)
{
	return (log(a) / log(2.0));
}


static double
myexp10(double a)
{
	return pow(10.0, a);
}

/*
 * Table of math functions
 *
 * NB: Sort in reverse to find log10 before log
 */

static struct functbl {
	const char *name;
	double (*func)(double);
} functbl[] = {
	{ "tan", 	tan },
	{ "sqrt", 	sqrt },
	{ "sin", 	sin },
	{ "round", 	round },
	{ "log2", 	mylog2 },
	{ "log10", 	log10 },
	{ "log", 	log },
	{ "floor", 	floor },
	{ "fabs", 	fabs },
	{ "exp2", 	exp2 },
	{ "exp10", 	myexp10 },
	{ "exp", 	exp },
	{ "cos", 	cos },
	{ "ceil", 	ceil },
	{ NULL, NULL }
};

/*
 * Primary units of expression, ie: parenthesis, numbers and functions
 */
static double
primary(struct e *e)
{
	double v;
	char *r;
	struct functbl *ft;

	dbg(e, "primary{");
	skip_space(e);
	if (e->p >= e->e || e->err != NULL) {
		return (NAN);
	} else if (*e->p == '(') {
		e->p++;
		v = expr(e, 4);
		skip_space(e);
		if (*e->p != ')') {
			e->err = "Missing ')'";
			return (0);
		}
		e->p++;
		skip_space(e);
	} else if (isalpha(*e->p)) {
		for(ft = functbl; ft->name != NULL; ft++) {
			if (e->p + strlen(ft->name) + 2 >= e->e)
				continue;
			if (!memcmp(e->p, ft->name, strlen(ft->name)))
				break;
		}
		if (ft->name == NULL) {
			e->err = "Unknown function";
			return (0);
		}
		e->p += strlen(ft->name);
		skip_space(e);
		if (*e->p != '(') {
			e->err = "expected '(' in function call";
			return (0);
		}
		v = primary(e);
		v = ft->func(v);
	} else {
		errno = 0;
		v = strtod(e->p, &r);
		if (r == e->p) {
			e->err = "Number expected";
			return (0);
		}
		e->p = r;
	}
	dbg(e, "primary}");
	return (v);
}

/*
 * Unary operations: + and -
 */
static double
unary(struct e *e)
{
	double v;

	dbg(e, "unary{");
	skip_space(e);
	if (e->p >= e->e || e->err != NULL) {
		return (NAN);
	} else if (*e->p == '+') {
		e->p++;
		v = unary(e);
	} else if (*e->p == '-') {
		e->p++;
		v = -unary(e);
	} else {
		v = primary(e);
	}
	dbg(e, "unary}");
	return (v);
}

/*
 * evaluate an expression above a certain level of precedence
 */
static double
expr(struct e *e, int level)
{
	double v, u, w;
	int k1;
	int oper;

	dbg(e, "expr{");
	v = unary(e);
	get_token(e);
	for (k1 = e->l; k1 >= level; k1--) {
		while (e->l == k1) {
			oper = e->t;
			u = expr(e, k1 + 1);
			switch (oper) {
			case '+':	w = v + u; break;
			case '-':	w = v - u; break;
			case '*':	w = v * u; break;
			case '/':	w = v / u; break;
			case '^':	w = pow(v, u); break;
			default:
				assert(oper == '?');
				w = 0;
			}
			v = w;
		}
	}
	dbg(e, "expr}");
	return (v);
}

/*
 * Exported function
 *
 * Returns NAN on error, providing explanation in errs (if non-null).
 * The error string must be passed to free(3)
 *
 */
double
phk_expr(const char *p, char **errs)
{
	struct e e[1];
	double v;

	memset(e, 0, sizeof e);
	if (errs != NULL)
		*errs = 0;
	e->b = p;
	e->p = p;
	e->e = strchr(p, '\0');
	assert(e->e != NULL);
	v = expr(e, 4);
	if (e->err == NULL)
		skip_space(e);
	if (e->err == NULL && e->p != e->e)
		e->err = "Syntax error";
	if (e->err == NULL && isnan(v))
		e->err = "Arithmetical error";
	if (e->err != NULL) {
		if (errs != NULL)
			asprintf(errs, "Expression error:\n%s\n%*.*s^ %s\n",
			    e->b, e->p - e->b, e->p - e->b, "", e->err);
		return (NAN);
	}
	return (v);
}
