/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: phk_ihex.c,v 1.2 2006/09/09 13:51:39 phk Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/queue.h>
#include <err.h>
#include <ctype.h>
#include <assert.h>
#include <printf.h>
#include <string.h>

#include "phk_ihex.h"

struct phkihex_line {
	unsigned				magic;
#define MAGICL 0xb9fe02f5
	TAILQ_ENTRY(phkihex_line)	next;
	unsigned				adr;
	unsigned				len;
	unsigned char				*data;
};

struct phkihex {
	unsigned				magic;
#define MAGIC 0x7798aba3
	TAILQ_HEAD(,phkihex_line)	list;
};

int
PhkIhexGet(struct phkihex **ihfp, unsigned adr, unsigned len, void *data)
{
	struct phkihex *ihf;
	struct phkihex_line *ihfl;
	unsigned char *p, *q, *r;
	unsigned l;
	int retval;

	assert(data != NULL);
	assert(ihfp != NULL);
	ihf = *ihfp;
	assert(ihf != NULL);
	assert(ihf->magic == MAGIC);
	p = data;
	retval = 0;
	TAILQ_FOREACH(ihfl, &ihf->list, next) {
		assert(ihfl->magic == MAGICL);
		if (ihfl->adr >= adr + len)
			continue;
		if (ihfl->adr + ihfl->len <= adr)
			continue;
		if (ihfl->adr < adr) {
			q = ihfl->data + (adr - ihfl->adr);
			l = ihfl->len - (adr - ihfl->adr);
			r = p;
		} else {
			q = ihfl->data;
			l = ihfl->len;
			r = &p[ihfl->adr - adr];
		}
		if (r + l > p + len)
			l -= ((r + l) - (p + len));
		retval += l;
		memcpy(r, q, l);
	}
	return (retval);
}


void
PhkIhexFree(struct phkihex **ihfp)
{
	struct phkihex *ihf;
	struct phkihex_line *ihfl;

	assert(ihfp != NULL);
	ihf = *ihfp;
	assert(ihf != NULL);
	assert(ihf->magic == MAGIC);

	*ihfp = NULL;

	while (!TAILQ_EMPTY(&ihf->list)) {
		ihfl = TAILQ_FIRST(&ihf->list);
		assert(ihf->magic == MAGICL);
		TAILQ_REMOVE(&ihf->list, ihfl, next);
		free(ihfl->data);
		ihfl->magic = 0;
		free(ihfl);
	}
	ihf->magic = 0;
	free(ihf);
}

int
PhkIhexRead(const char *filename, struct phkihex **ihfp)
{
	struct phkihex *ihf;
	struct phkihex_line *ihfl;
	FILE *fi;
	char buf[BUFSIZ];
	unsigned char hbuf[BUFSIZ], *up, cc;
	unsigned u, v;
	int i;
	unsigned ll, aaaa, tt;
	unsigned uadr;
	int debug = 0;

	assert(ihfp != NULL);

	if (debug)
		register_printf_render_std("VH");

	fi = fopen(filename, "r");
	if (fi == NULL) {
		warnx("Cannot open %s\n", filename);
		return (-1);
	}

	ihf = calloc(sizeof *ihf, 1);
	if (ihf == NULL) {
		warnx("Cannot malloc\n");
		fclose(fi);
		return (-1);
	}

	ihf->magic = MAGIC;

	TAILQ_INIT(&ihf->list);
	*ihfp = ihf;

	uadr = 0;
	while (fgets(buf, sizeof buf, fi) != NULL) {
		if (*buf != ':') {
			warnx("Input file had no colon");
			fclose(fi);
			return (-1);
		}
		up = hbuf;
		for (u = 1; isxdigit(buf[u]) && isxdigit(buf[u + 1]); u += 2) {
			i = sscanf(buf + u, "%02x", &v);
			assert(i == 1);
			*up++ = v;
		}
		if (buf[u] != '\n' && buf[u] != '\r') {
			warnx("Input file end of line not NL or CR");
			fclose(fi);
			return (-1);
		}

		cc = 0;
		for (u = 0; &hbuf[u] != up; u++)
			cc += hbuf[u];
		ll = hbuf[0];
		aaaa = (hbuf[1] << 8) | hbuf[2];
		tt = hbuf[3];

		switch (tt) {
		case 0x04:
			uadr = hbuf[4] << 24;
			uadr |= hbuf[5] << 16;
			break;
		case 0x00:
			ihfl = calloc(sizeof *ihfl, 1);
			if (ihfl == NULL) {
				warnx("Cannot malloc");
				fclose(fi);
				return (-1);
			}
			ihfl->magic = MAGICL;
			ihfl->data = calloc(ll, 1);
			if (ihfl->data == NULL) {
				warnx("Cannot malloc");
				fclose(fi);
				return (-1);
			}
			memcpy(ihfl->data, hbuf + 4, ll);
			ihfl->adr = aaaa | uadr;
			ihfl->len = ll;
			TAILQ_INSERT_TAIL(&ihf->list, ihfl, next);
			break;
		case 0x01:
			break;
		default:
			warnx("Unhandled iHEX record: %02x", tt);
			fclose(fi);
			return (-1);
		}
	}
	fclose(fi);

	return (0);
}

#if 0
#include <stdio.h>
#include <string.h>

#include "ihex.h"

int
main(int argc __unused, char **argv __unused)
{
	struct phkihex *ihf;
	int i;
	unsigned char hbuf[BUFSIZ];

	i = ReadIntelHexFile("/home/phk/Sw/ICSP/t.hex", &ihf);
	printf("i = %d\n", i);

	memset(hbuf, 0x55, sizeof hbuf);
	i = GetIntelHex(&ihf, 0x0, sizeof hbuf, hbuf);
	printf("i = %d \n%+H\n", i, hbuf, sizeof hbuf);

	FreeIntelHexFile(&ihf);

#if 0
	i = ReadIntelHexFile("/backup/Tmp/040441-11/DS18S20/DS1820.hex", &ihf);
	printf("i = %d\n", i);

	memset(hbuf, 0x55, sizeof hbuf);
	i = GetIntelHex(&ihf, 0x0, sizeof hbuf, hbuf);
	printf("i = %d \n%+H\n", i, hbuf, sizeof hbuf);

	FreeIntelHexFile(&ihf);
#endif
	return (0);
}
#endif

/*--------------------------------------------------------------------*/
